(* syntax.sml *)

structure Syntax =
struct

local
open Error
in


(* ---------------------------------------------------------------------- *)
(* Datatypes *)

type info = Error.info

datatype ty
  = TyBool
  | TyArr of ty * ty

datatype term
  = TmTrue of info
  | TmFalse of info
  | TmIf of info * term * term * term
  | TmVar of info * int * int
  | TmAbs of info * string * ty * term
  | TmApp of info * term * term

datatype binding
  = NameBind       (* dummy binding used only during parsing *)
  | VarBind of ty

type context = (string * binding) list

datatype command
  = Eval of info * term
  | Bind of info * string * binding

(* equality of types *)
fun equalTy (TyBool, TyBool) = true
  | equalTy (TyArr(t1,t2),TyArr(u1,u2)) =
     equalTy(t1,u1) andalso equalTy(t2,u2)
  | equalTy _ = false

fun tmInfo(TmVar(fi,_,_)) = fi
  | tmInfo(TmAbs(fi,_,_,_)) = fi
  | tmInfo(TmApp(fi,_,_)) = fi
  | tmInfo(TmTrue fi) = fi
  | tmInfo(TmFalse fi) = fi
  | tmInfo(TmIf(fi,_,_,_)) = fi


(* ---------------------------------------------------------------------- *)
(* Context management *)

val emptycontext = []

fun ctxlength (ctx: context) = List.length ctx

fun addbinding (ctx: context, x: string, bind) = (x,bind)::ctx

fun addname(ctx,x) = addbinding(ctx,x,NameBind)

fun isnamebound (ctx: context, x: string) =
    case ctx
      of [] => false
       | (y,_)::rest =>
          if y=x then true
          else isnamebound(rest,x)

fun pickfreshname (ctx,x) =
    if isnamebound(ctx,x) then pickfreshname(ctx,(x^"'"))
    else ((x,NameBind)::ctx, x)

fun index2name (ctx: context,x) =
    #1(List.nth(ctx,x))
    handle Size =>
      let val msg =
	      concat["Variable lookup failure: offset: ", Int.toString x,
		     ", ctx size: ", Int.toString (List.length ctx)]
       in error NONE msg
      end

fun name2index(ctx,x) =
    case ctx
      of [] => error NONE ("Identifier " ^ x ^ " is unbound")
       | (y,_)::rest =>
          if y = x then 0
          else 1 + name2index(rest,x)

fun getbinding(fi,ctx:context,i) =
    #2(List.nth(ctx,i))
    handle Size =>
      let val msg =
	      concat["Variable lookup failure: offset: ", Int.toString i,
		     ", ctx size: ", Int.toString (List.length ctx)]
       in error fi msg
      end

fun getTypeFromContext(fi,ctx,i) =
    case getbinding(fi,ctx,i)
      of VarBind(tyT) => tyT
       | _ => error fi
            ("getTypeFromContext: Wrong kind of binding for variable " 
             ^ index2name(ctx,i)) 


(* ---------------------------------------------------------------------- *)
(* Shifting *)

fun tmmap onvar c t =
    let fun walk(c,t) =
            case t
	      of TmTrue _ => t
	       | TmFalse _ => t
	       | TmIf(fi,t1,t2,t3) => TmIf(fi,walk(c,t1),walk(c,t2),walk(c,t3))
	       | TmVar(fi,x,n) => onvar(fi,c,x,n)
	       | TmAbs(fi,x,tyT1,t2) => TmAbs(fi,x,tyT1,walk(c+1,t2))
	       | TmApp(fi,t1,t2) => TmApp(fi,walk(c,t1),walk(c,t2))
     in walk(c,t)
    end

fun termShiftAbove(d,c,t) =
  tmmap
    (fn (fi,c,x,n) => if x>=c then TmVar(fi,x+d,n+d) else TmVar(fi,x,n+d))
    c t

fun termShift(d,t) = termShiftAbove(d,0,t)

(* ---------------------------------------------------------------------- *)
(* Substitution *)

fun termSubst(j,s,t) =
  tmmap
    (fn (fi,j,x,n) => if x=j then termShift(j,s) else TmVar(fi,x,n))
    j t

fun termSubstTop(s,t) = 
    termShift(~1, termSubst(0,termShift(1,s),t))
(* ---------------------------------------------------------------------- *)
(* Printing *)

(* The printing functions call these utility functions to insert grouping
  information and line-breaking hints for the pretty-printing library:
     obox   Open a "box" whose contents will be indented by two spaces if
            the whole box cannot fit on the current line
     obox0  Same but indent continuation lines to the same column as the
            beginning of the box rather than 2 more columns to the right
     cbox   Close the current box
     break  Insert a breakpoint indicating where the line maybe broken if
            necessary.
  See the documentation for the Format module in the OCaml library for
  more details. 
*)

open PPUtil

fun obox0() = open_hvbox 0
fun obox() = open_hvbox 2
fun cbox() = close_box()
fun break() = print_break 0 0
val pr = print_string

fun small t = 
    case t
      of TmVar(_,_,_) => true
       | _ => false

fun printty_Type(outer,tyT) =
    printty_ArrowType(outer,tyT)

and printty_ArrowType(outer,tyT) =
    case tyT
      of TyArr(tyT1,tyT2) =>
	 (obox0();
	  printty_AType(false,tyT1);
	  if outer then pr " " else  ();
	  pr "->";
	  if outer then print_space() else break();
	  printty_ArrowType(outer,tyT2);
	  cbox())
       | _ => printty_AType(outer,tyT)

and printty_AType(outer,tyT) =
    case tyT
      of TyBool => pr "Bool"
       | tyT => (pr "("; printty_Type(outer,tyT); pr ")")

fun printty tyT = printty_Type(true,tyT)

fun printtm_Term(outer,ctx,t) =
    case t
      of TmIf(_, t1, t2, t3) =>
       (obox0();
	pr "if ";
	printtm_Term(false,ctx,t1);
	print_space();
	pr "then ";
	printtm_Term(false,ctx,t2);
	print_space();
	pr "else ";
	printtm_Term(false,ctx,t3);
	cbox())
  | TmAbs(_,x,tyT1,t2) =>
       let val (ctx',x') = pickfreshname(ctx,x)
        in obox(); pr "lambda ";
           pr x'; pr ":"; printty_Type(false,tyT1); pr ".";
           if (small t2) andalso not outer then break() else print_space();
           printtm_Term(outer,ctx',t2);
           cbox()
       end
  | t => printtm_AppTerm(outer,ctx,t)

and printtm_AppTerm(outer,ctx,t) =
    case t
      of TmApp(fi, t1, t2) =>
         (obox0();
          printtm_AppTerm(false,ctx,t1);
          print_space();
          printtm_ATerm(false,ctx,t2);
          cbox())
       | _ => printtm_ATerm(outer,ctx,t)

and printtm_ATerm(outer,ctx,t) =
    case t
      of TmTrue _ => pr "true"
       | TmFalse _ => pr "false"
       | TmVar(_,x,n) =>
	  if ctxlength ctx = n then
	    pr (index2name(ctx,x))
	  else
	    pr ("[bad index: " ^ (Int.toString x) ^ "/" ^ (Int.toString n)
		^ " in {"
		^ (List.foldl (fn ((x,_),s) => s ^ " " ^ x) "" ctx)
		^ " }]")
       | t => (pr "("; printtm_Term(outer,ctx,t); pr ")")

fun printtm(ctx,t) = printtm_Term(true,ctx,t)

fun prbinding(ctx,b) =
    case b
      of NameBind => ()
       | VarBind(tyT) => (pr ": "; printty tyT)


end (* local *)
end (* structure Syntax *)
